// aa-
// Proteus (lawnmower-man@mail.ru)

#include<stdio.h>
#include<stdlib.h>

class AA_tree
{
protected:
	struct node_st{
		int value;	//!< 
		int level;	//!<    
		node_st *p1,*p2;
	//
		node_st(int v):level(0),p1(0),p2(0),value(v) {}
		int Level() {return this?level:-1;}
	};
	node_st *tree_root;	//!<  
private:
	int count;			//!<  
	void Clear(node_st*);
	node_st *Insert(node_st*,int v);
	node_st *Remove(node_st*,int v);
	node_st *Skew(node_st*);
	node_st *Split(node_st*);
public:
	AA_tree():count(0),tree_root(0) {}
	~AA_tree()         {Clear(tree_root);}
	void Insert(int v) {tree_root=Insert(tree_root,v);}
	void Remove(int v) {tree_root=Remove(tree_root,v);}
	int  Size()        {return count;}
};


void AA_tree::Clear(node_st *node)
{
	if(!node) Clear(node->p1),Clear(node->p2);
	delete node;
}


AA_tree::node_st *AA_tree::Insert(node_st *node,int v)
{
	if(!node) {count++;return new node_st(v);}
	if(v<node->value)      node->p1=Insert(node->p1,v);
	else if(v>node->value) node->p2=Insert(node->p2,v);
	else return node;
	return Split(Skew(node));
}


AA_tree::node_st *AA_tree::Remove(node_st *node,int v)
{
	static node_st *last_node, *del_node=0;
	if(node)
	{
		// Step 1
		last_node=node;
		if(v<node->value)  node->p1=Remove(node->p1,v);
		else del_node=node,node->p2=Remove(node->p2,v);
		// Step 2
		if(node==last_node)
		{
			if(!del_node || del_node->value!=v) return node;
			del_node->value=node->value,del_node=0;
			node=node->p2;
			delete last_node; count--;
		}
		// Step 3
		else if(node->p1->Level()<node->level-1 || node->p2->Level()<node->level-1) {
			 if(node->p2->Level()>--node->level) node->p2->level=node->level;
				node=Skew(node);
				node->p2=Skew(node->p2);
				if(node->p2->p2) node->p2->p2=Skew(node->p2->p2);
				node=Split(node);
				node->p2=Split(node->p2);
			}
	}
	return node;
}


AA_tree::node_st *AA_tree::Skew(node_st *node)
{
	node_st *p1=node->p1;
	if(p1 && p1->level==node->level) node->p1=p1->p2,p1->p2=node,node=p1;
	return node;
}


AA_tree::node_st *AA_tree::Split(node_st *node)
{
	node_st *p2=node->p2;
	if(p2 && p2->p2->Level()==node->level) node->p2=p2->p1,p2->p1=node,node=p2,node->level++;
	return node;
}

//!  
class TreeTest:public AA_tree
{
	int  Test(node_st*);
	void Show(node_st*,int);
public:
	void Show() {Show(tree_root,0);}
	int  Test() {return Test(tree_root);}
};

//!     2-3 
void TreeTest::Show(node_st *node,int ident)
{
	if(!node) return;
	node_st *i; 
	printf("%*s",ident++*3,"");
	for(i=node;i && i->Level()==node->level;i=i->p2) printf(" %d:[%d]",i->value,i->level); putchar('\n');
	for(i=node;i && i->Level()==node->level;i=i->p2) Show(i->p1,ident);
	if(i) Show(i,ident);
}

//!  
//! \result    -1   
int TreeTest::Test(node_st *node)
{
	if(!node) return 0;
	int n; node_st *i;
	node_st *t[4];      //   2-3 
	int v[4]; int vi=0; //  2-3 
	int d[4]; int ti=0; //  
	for(i=node;i && i->Level()==node->level ;i=i->p2) v[vi++]=i->value;
	for(i=node;i && i->Level()==node->level ;i=i->p2) if(i->p1) t[ti++]=i->p1;
	if(i) t[ti++]=i;
	//   2-3 
	if(!(ti==0 && vi==1 || ti==0 && vi==2 || ti==2 && vi==1 || ti==3 && vi==2)) return -1;
	if(!ti) return 1;
	for(n=0; n<ti; n++) if((d[n]=Test(t[n]))<0) return -1;
	for(n=1; n<ti; n++) if(d[0]!=d[n]) return -1;
	return d[0]+1;
}

	
#define SIZE 10000

int main( )
{
	int n,v;
  	TreeTest tree;
	for(n=0; n<30; n++) tree.Insert(rand()%100);
	tree.Show(); // 
	getchar(); 
	//  -----------------------------
	for(n=0;n<SIZE*10;n++) {
		v=rand()%SIZE; printf("[ins: %5d]",v); tree.Insert(v);
		v=rand()%SIZE; printf("[del: %5d]",v); tree.Remove(v);
		printf("[size: %d]\n",tree.Size());
		if(tree.Test()<0) {printf("Internal error!!\n\a");}
	}
	system("pause");
    return 0;
}

